%if style = code
\begin{code}
{-# OPTIONS --no-termination-check #-}
{-# OPTIONS --no-positivity-check #-}
  module zipperAux where

    open import appendix
    --open import Relation.Binary.PropositionalEquality public hiding ([_])
    --open import Coinduction public
    open import Data.Empty public
    open import Data.Unit using (⊤ ; tt)
    open import Data.Sum public hiding (map)
    open import Data.Product public hiding (map ; ∃)
    open import Function public

    infixr 3 _⊕_
    infixr 4 _⊗_
    infixr 5 _⊚_

    mutual
\end{code}

\begin{code}
      data _▸_ : Set -> Set -> Set₁ where
        Z'      :  ∀ {I O} -> I ▸ O
        U'      :  ∀ {I O} -> I ▸ O

        I'     :  ∀ {I O} -> I  -> I ▸ O
        !      :  ∀ {I O} -> O  -> I ▸ O

        _⊕_    :  ∀ {I O}     -> I ▸ O  -> I ▸ O  ->  I ▸ O
        _⊗_    :  ∀ {I O}     -> I ▸ O  -> I ▸ O  ->  I ▸ O
        _⊚_    :  ∀ {I M O}   -> M ▸ O  -> I ▸ M  ->  I ▸ O

        _↗_↘_  :  ∀ {I I′ O O′} -> I′ ▸ O′
               ->  (I′ -> I) -> (O -> O′) ->  I ▸ O

        Fix'    :  ∀ {I O} -> (I ⊎ O) ▸ O -> I ▸ O

        Sig      :  ∀ {I O}  -> {C : ⊥ ▸ ⊤} 
               ->  (⟦ C ⟧ (λ _ -> ⊤) tt -> I ▸ O) -> I ▸ O

        Iso    :  ∀ {I O}  -> (C : I ▸ O) -> (D : I ▷ O)
               ->  ((r : Indexed I) -> (o : O) -> D r o ≃ ⟦ C ⟧ r o)
               ->  I ▸ O
      
      Code : Set -> Set -> Set₁
      Code = _▸_
\end{code}
\begin{code}
      data μ  {I O : Set} (F : (I ⊎ O) ▸ O) 
              (r : Indexed I) (o : O) : Set where
        ⟨_⟩ : ⟦ F ⟧ (r ∣ μ F r) o -> μ F r o
    
      ⟦_⟧ : ∀ {I O} -> I ▸ O -> I ▷ O
      ⟦ Z'     ⟧      r o =  ⊥
      ⟦ U'     ⟧      r o =  ⊤
      ⟦ I' i  ⟧      r o =  r i
      ⟦ F ↗ f ↘ g ⟧  r o =  ⟦ F ⟧ (r ∘ f) (g o)
      ⟦ F ⊕ G ⟧      r o =  ⟦ F ⟧ r o ⊎ ⟦ G ⟧ r o
      ⟦ F ⊗ G ⟧      r o =  ⟦ F ⟧ r o × ⟦ G ⟧ r o
      ⟦ F ⊚ G ⟧      r o =  ⟦ F ⟧ (⟦ G ⟧ r) o
      ⟦ Fix' F ⟧      r o =  μ F r o
      ⟦ ! o′  ⟧      r o =  o ≡ o′
      ⟦ Sig f   ⟧      r o =  ∃ (λ i -> ⟦ f i ⟧ r o)
      ⟦ Iso C D e ⟧  r o =  D r o
\end{code}
\begin{code}
      map  :  {I O : Set} {r s : Indexed I} (C : I ▸ O)
           ->  (r ⇉ s) -> (⟦ C ⟧ r ⇉ ⟦ C ⟧ s)
      map Z'            f o ()
      map U'            f o x                   = x
      map (I' i)       f o x                   = f i x
      map (F ⊕ G)      f o (inj₁ x)             = inj₁ (map F f o x)
      map (F ⊕ G)      f o (inj₂ y)             = inj₂ (map G f o y)
      map (F ⊗ G)      f o (x , y)             = map F f o x , map G f o y
      map (F ↗ g ↘ h)  f o x                   = map F (f ∘ g) (h o) x
      map (F ⊚ G)      f o x                   = map F (map G f) o x
      map (! o')       f o x                   = x
      map (Sig g)        f o (some {i} x)        = some (map (g i) f o x)

      map (Fix' F)      f o ⟨ x ⟩ = ⟨ map F (f ∥ map (Fix' F) f) o x ⟩

      map {r = r} {s = s} (Iso C D e) f o x with (e r o , e s o)
      ... | (ep₁ , ep₂) = _≃_.to ep₂ (map C f o (_≃_.from ep₁ x))
\end{code}

\begin{code}
    T : Set -> Indexed ⊤
    T A tt = A

    data Maybe (A : Set) : Set where
      nothing : Maybe A
      just    : A -> Maybe A

    mapMaybe : {A B : Set} -> (A -> B) -> Maybe A -> Maybe B
    mapMaybe f nothing  = nothing
    mapMaybe f (just x) = just (f x)

    infixl 5 _>>=_
    _>>=_ : {A B : Set} -> Maybe A -> (A -> Maybe B) -> Maybe B
    nothing >>= f = nothing
    just x  >>= f = f x

    plusMaybe : {A : Set} -> Maybe A -> Maybe A -> Maybe A
    plusMaybe (just x) _ = just x
    plusMaybe nothing  y = y

    infixl 6 _>=>_
    _>=>_ : {A B C : Set} -> (A -> Maybe B) -> (B -> Maybe C) -> A -> Maybe C
    (f >=> g) x with f x
    (f >=> g) x | nothing = nothing
    (f >=> g) x | just y  = g y

    data Nat : Set where
      ze : Nat
      su : Nat -> Nat



    `NatF' : (⊥ ⊎ ⊤) ▸ ⊤
    `NatF' = U' ⊕ I' (inj₂ tt)

    fromNat : ∀ {r o} -> Nat -> ⟦ Fix' `NatF' ⟧ r o
    fromNat ze      = ⟨ inj₁ tt ⟩
    fromNat (su n)   = ⟨ inj₂ (fromNat n) ⟩

    toNat : ∀ {r o} -> ⟦ Fix' `NatF' ⟧ r o -> Nat
    toNat ⟨ inj₁ tt ⟩  = ze
    toNat ⟨ inj₂ n ⟩   = su (toNat n)

    toFromNat≃ : ∀ {r o} (n : Nat) -> toNat {r} {o} (fromNat n) ≡ n
    toFromNat≃ ze  = refl
    toFromNat≃ (su _) = cong su (toFromNat≃ _)

    fromToNat≃ : ∀ {r o} (x : ⟦ Fix' `NatF' ⟧ r o) -> (fromNat (toNat x)) ≡ x
    fromToNat≃ ⟨ inj₁ tt ⟩ = refl
    fromToNat≃ ⟨ inj₂ _  ⟩ = cong (⟨_⟩ ∘ inj₂) (fromToNat≃ _)

    ‵ℕ' : ⊥ ▸ ⊤
    ‵ℕ' = Iso (Fix' `NatF') (λ _ _ -> Nat) (λ r o -> record { from = fromNat 
                                                     ; to   = toNat 
                                                     ; iso₁ = toFromNat≃
                                                     ; iso₂ = fromToNat≃ })

    data Perfect (A : Set) : {n : Nat} -> Set where
      split : {n : Nat} -> A -> Perfect A {n} × Perfect A {n} -> Perfect A {su n}
      leaf  : Perfect A {ze}

    PerfectF : Nat -> (⊤ ⊎ Nat) ▸ Nat
    PerfectF (ze)    = !  ze
    PerfectF (su n)   = !  (su n) ⊗ I' (inj₁ tt) ⊗ I' (inj₂ n) ⊗ I' (inj₂ n)

    `PerfectF' : (⊤ ⊎ Nat) ▸ Nat
    `PerfectF' = Sig {C = ‵ℕ'} PerfectF

    `Perfect' : ⊤ ▸ Nat
    `Perfect' = Fix' `PerfectF'

    fromPerfect′ : {r : Indexed ⊤} {n : Nat} -> Perfect (r tt) {n} -> ⟦ `Perfect' ⟧ r n
    fromPerfect′ {n = su n} (split x (p1 , p2)) = 
      ⟨ some {x = su n} (refl , (x , (fromPerfect′ p1 , fromPerfect′ p2))) ⟩
    fromPerfect′ {n = ze} leaf = ⟨ some {x = ze} refl ⟩

    fromPerfect : {A : Set} {n : Nat} -> Perfect A {n} -> ⟦ `Perfect' ⟧ (T A) n
    fromPerfect = fromPerfect′

    toPerfect′ : {r : Indexed ⊤} {n : Nat} -> ⟦ `Perfect' ⟧ r n -> Perfect (r tt) {n}
    toPerfect′ ⟨ some {ze}  refl ⟩                     = leaf
    toPerfect′ ⟨ some {su n} (refl , (x , (p1 , p2))) ⟩ =
      split x (toPerfect′ p1 , toPerfect′ p2)

    toPerfect : {A : Set} {n : Nat} -> ⟦ `Perfect' ⟧ (T A) n -> Perfect A {n}
    toPerfect = toPerfect′
{-
    `PerfectE' : ⊤ ▸ Nat
    `PerfectE' = Iso `Perfect' (λ f n -> Perfect (f tt) {n})
                              (λ r n -> record { from = fromPerfect′ 
                                              ; to   = toPerfect′
                                              ; iso₁ = nothing ; iso₂ = nothing })

    mapPerfect : ∀ {n A B} -> (A -> B) -> Perfect A {n} -> Perfect B {n}
    mapPerfect {n} f = map `PerfectE' (↑ f) n
-}
    id⇉ : {I : Set} {r : Indexed I} -> r ⇉ r
    id⇉ i = id

    cata  : {I O : Set} {r : Indexed I} {s : Indexed O}
          -> (C : (I ⊎ O) ▸ O) -> (⟦ C ⟧ (r ∣ s) ⇉ s) -> ⟦ Fix' C ⟧ r ⇉ s
    cata  C  φ     o  ⟨ x ⟩  = φ o (map C (id⇉ ∥ cata C φ) o x)

    cataPerfect : {n : Nat} {A : Set} {R : Nat -> Set}
                -> R ze
                -> ({m : Nat} -> A -> R m × R m -> R (su m))
                -> Perfect A {n} -> R n
    cataPerfect {n} {A} {R} r f p = cata `PerfectF' g n (fromPerfect p)
      where
        g : (i : Nat) -> ∃ (λ i′ -> ⟦ PerfectF i′ ⟧ (T A ∣ R) i) -> R i
        g .ze     (some {ze}   refl)                     = r
        g .(su n) (some {su n} (refl , (x , (p1 , p2)))) = f x (p1 , p2)

    cataPerfect' : ∀ {n A R} -> R -> (A -> R × R -> R) -> Perfect A {n} -> R
    cataPerfect' {R = R} r f p = cataPerfect {R = const R} r f p


    mutual
      data Expr  (A : Set)  : Set where
        econst  : Nat                 -> Expr A
        add     : Expr A  -> Expr A  -> Expr A
        evar    : A                 -> Expr A
        elet    : Decl A  -> Expr A  -> Expr A

      data Decl  (A : Set)  : Set where
        assign  : A       -> Expr A  -> Decl A
        seq     : Decl A  -> Decl A  -> Decl A

    AST : Set
    AST = ⊤ ⊎ ⊤

    expr  : AST
    expr  = inj₁ tt

    decl  : AST
    decl  = inj₂ tt

    `ExprF' : (⊤ ⊎ AST) ▸ AST
    `ExprF'  =     ‵ℕ' ↗ (λ ()) ↘ (const tt)
                ⊕  I' (inj₂  expr)  ⊗ I' (inj₂ expr)
                ⊕  I' (inj₁  tt)
                ⊕  I' (inj₂  decl)  ⊗ I' (inj₂ expr)

    `DeclF' : (⊤ ⊎ AST) ▸ AST
    `DeclF'  =     I' (inj₁  tt)    ⊗ I' (inj₂ expr)
                ⊕  I' (inj₂  decl)  ⊗ I' (inj₂ decl)

    `ASTF'  : (⊤ ⊎ AST) ▸ AST
    `ASTF'  =     !  expr  ⊗  `ExprF'
               ⊕  !  decl  ⊗  `DeclF'

    `AST' : ⊤ ▸ AST
    `AST' = Fix' `ASTF'

    mutual
      toExpr : {r : ⊤ -> Set} -> ⟦ `AST' ⟧ r expr -> Expr (r tt)
      toExpr ⟨ inj₁  (refl  , inj₁ x) ⟩                    = econst x
      toExpr ⟨ inj₁  (refl  , inj₂ (inj₁ (x , y))) ⟩        = add (toExpr x) (toExpr y)
      toExpr ⟨ inj₁  (refl  , inj₂ (inj₂ (inj₁ x))) ⟩        = evar x
      toExpr ⟨ inj₁  (refl  , inj₂ (inj₂ (inj₂ (d , e)))) ⟩  = elet (toDecl d) (toExpr e)
      toExpr ⟨ inj₂  (()    , _) ⟩

      toDecl : {r : ⊤ -> Set} -> ⟦ `AST' ⟧ r decl -> Decl (r tt)
      toDecl ⟨  inj₁ (()    , _) ⟩
      toDecl ⟨  inj₂ (refl  , inj₁ (x , e)) ⟩    = assign x (toExpr e)
      toDecl ⟨  inj₂ (refl  , inj₂ (d₁ , d₂)) ⟩  = seq (toDecl d₁) (toDecl d₂)

    mutual
      fromExpr : {r : ⊤ -> Set} -> Expr (r tt) -> ⟦ `AST' ⟧ r expr
      fromExpr (econst x)  = ⟨ inj₁ (refl , inj₁ x) ⟩
      fromExpr (add x y)   = ⟨ inj₁ (refl , inj₂ (inj₁ (fromExpr x , fromExpr y))) ⟩
      fromExpr (evar x)    = ⟨ inj₁ (refl , inj₂ (inj₂ (inj₁ x))) ⟩
      fromExpr (elet d e)  = ⟨ inj₁ (refl , inj₂ (inj₂ (inj₂ (fromDecl d , fromExpr e))))⟩

      fromDecl : {r : ⊤ -> Set} -> Decl (r tt) -> ⟦ `AST' ⟧ r decl
      fromDecl (assign x e)  = ⟨ inj₂ (refl , inj₁ (x , fromExpr e)) ⟩
      fromDecl (seq d₁ d₂)   = ⟨ inj₂ (refl , inj₂ (fromDecl d₁ , fromDecl d₂)) ⟩


    `ListF' : (⊤ ⊎ ⊤) ▸ ⊤
    `ListF' = U' ⊕ (I' (inj₁ tt) ⊗ I' (inj₂ tt))

    `List' : ⊤ ▸ ⊤
    `List' = Fix' `ListF'

    infixr 6 _∷_
    data [_] (A : Set) : Set where
      []  : [ A ]
      _∷_ : A -> [ A ] -> [ A ]

    fromList′ : {r : Indexed ⊤} {o : ⊤} -> [ r o ] -> ⟦ `List' ⟧ r o
    fromList′ {o = tt} []       = ⟨ inj₁ tt ⟩
    fromList′ {o = tt} (x ∷ xs) = ⟨ inj₂ (x , fromList′ xs) ⟩

    fromList : ∀ {A} -> [ A ] -> ⟦ `List' ⟧ (const A) tt
    fromList = fromList′

    toList′ : {r : Indexed ⊤} {o : ⊤} -> ⟦ `List' ⟧ r o -> [ r o ]
    toList′ {o = tt} ⟨ inj₁ tt ⟩       = []
    toList′ {o = tt} ⟨ inj₂ (x , xs) ⟩ = x ∷ toList′ xs

    toList : ∀ {A} -> ⟦ `List' ⟧ (const A) tt -> [ A ]
    toList = toList′

    isoList₁ : {r : Indexed ⊤} {o : ⊤} (x : [ r o ])
             -> toList′ {r} (fromList′ x) ≡ x
    isoList₁     []      = refl
    isoList₁ {r} (h ∷ t) = cong (_∷_ h) (isoList₁ {r} t)

    postulate isoList₂ : {r : Indexed ⊤} {o : ⊤} (x : ⟦ `List' ⟧ r o)
                       -> fromList′ {r} (toList′ x) ≡ x

    epList : {r : Indexed ⊤} {o : ⊤} -> [ r o ] ≃ ⟦ `List' ⟧ r o
    epList {r} = record { from = fromList′ 
                        ; to   = toList′
                        ; iso₁ = isoList₁ {r} ; iso₂ = isoList₂ }

    ‵List' : ⊤ ▸ ⊤
    ‵List' = Iso `List' (λ f t -> [ f t ]) (λ r o -> epList)


    `RoseF' : (⊤ ⊎ ⊤) ▸ ⊤
    `RoseF' = I' (inj₁ tt) ⊗ `List' ⊚ I' (inj₂ tt)

    ‵Rose′ : ⊤ ▸ ⊤
    ‵Rose′ = Fix' `RoseF'

    data Rose (A : Set) : Set where
      fork : A -> [ Rose A ] -> Rose A

    ↑ : ∀ {A B : Set} -> (A -> B) -> (_⇉_ {⊤} (const A) (const B))
    ↑ f i x = f x

    fromRose′ : {r : Indexed ⊤} {o : ⊤} -> Rose (r o) -> ⟦ ‵Rose′ ⟧ r o
    fromRose′ {o = tt} (fork x xs) = ⟨ x , fromList′ (map ‵List' (↑ fromRose′) tt xs) ⟩
    

    fromRose : {A : Set} -> Rose A -> ⟦ ‵Rose′ ⟧ (const A) tt
    fromRose = fromRose′


    toRose′ : {r : Indexed ⊤} {o : ⊤} -> ⟦ ‵Rose′ ⟧ r o -> Rose (r o)
    toRose′ {o = tt} ⟨ x , xs ⟩ = fork x (map ‵List' (↑ toRose′) tt (toList′ xs))

    toRose : {A : Set} -> ⟦ ‵Rose′ ⟧ (const A) tt -> Rose A
    toRose = toRose′
\end{code}
%endif
